/**
 *
 * Mozilla Public License Notice
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1  (the "License"); you may  not use this  file except in
 * compliance with the  License. You may obtain a  copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the  License  for  the   specific  language  governing  rights  and
 * limitations under the License.
 *
 * The  Original Code and  Idea is  the Hit-a-Hint  Mozilla extension.
 * The Initial  Developer of the Original  Code and the  Idea is Pekka
 * P. Sillanpaa.  Portions created  by Initial Developer are Copyright
 * (C) 2004.  All Rights Reserved.
 *
 * Contributor(s): Pekka Sillanpaa, Paul Stone
 *
 */

var hintElemArray = null; // array of hint elements
var linkElemArray = null;
var linkNumString = ""; // the typed link number is in this string
var hahQuickDisabled = false;
var isHahModeEnabled = false; // is typing mode on
var regenElemList = false;
var isFirstTime = true;
var areCoordsValid = true;
var hahDocCount = 0; // counter for frames in the document
var hahFrameArray = null;
var hahFAYTState = null;

var isMagicKeyPressed = false;

// not (yet?) in preferences
var BACKSPACE_KEY = 8; // backspace
var FINISH_KEY = 13; // enter
var STOP_KEY = 27; // esc

var HINT_PREFIX = 'hint'; // prefix for the hint id

////////////////////////////////////////////////////////////////////////////////
// logging
////////////////////////////////////////////////////////////////////////////////

var gConsoleService = Components.classes['@mozilla.org/consoleservice;1']
                    .getService(Components.interfaces.nsIConsoleService);

function hahLogMessage(aMessage)
{
  gConsoleService.logStringMessage('hah: ' + aMessage);
}

////////////////////////////////////////////////////////////////////////////////
// configuration and initialization related functions
////////////////////////////////////////////////////////////////////////////////

function hahLoad()
{
  if (isFirstTime) {
    hahPrefs.setDefaultPreferences();
    isFirstTime = false;
  }
  if(!hahPrefs.get(HAH_PREFS.ENABLED) || !hahPrefs.get(HAH_PREFS.STATUSICON_ENABLED)) {
    hahStatusIcon('', false);
  } else {
    hahStatusIcon("enabled", true);
  } 
  

  if(!hahQuickDisabled) {
    window.addEventListener("keydown", hahOnKeydown, true);
    window.addEventListener("keyup", hahOnKeyup, true);
    window.addEventListener("keypress", hahOnKeypress, true);
  }

  isHahModeEnabled = false;

  //window._content.scrollbars.visible = false;
}

function hahBlurFocused() {
  var elt = document.commandDispatcher.focusedElement;
  if (elt != null) elt.blur();
  //  getDocument().body.focus();
}

function hahResize() 
{
  areCoordsValid = false;
  if (isHahModeEnabled) {
    hahRemoveHints();
    hahShowHints();
  }
}

function hahQuickToggle() 
{
  if (hahQuickDisabled) {
    hahQuickDisabled = false;
    hahLoad();
    hahStatusIcon("enabled", true);
  }
  else {
    hahQuickDisabled = true;
    
    window.removeEventListener("keydown", hahOnKeydown, true);
    window.removeEventListener("keyup", hahOnKeyup, true);
    window.removeEventListener("keypress", hahOnKeypress, true);
  
    hahStatusIcon("disabled", true);
  }
}

////////////////////////////////////////////////////////////////////////////////
// hint activating and loading related functions
////////////////////////////////////////////////////////////////////////////////

function getNextWin()
{
  //var hahFrameArray = window._content.document.getElementsByTagName('FRAME');

//  hahLogMessage(hahDocCount);

  if (hahFrameArray.length > 0 && hahDocCount < hahFrameArray.length) {
    return hahFrameArray[hahDocCount++];
  }
  return false;
}

function hahShowHints()
{
  var counter = 0;

  hahStatusIcon("wait", true); // why doesn't work? xul don't refresh?

  hahDocCount = 0;

  linkElemArray = new Array();
  hintElemArray = new Array();
  var area = new Array(4);
  var win = window._content;

  // win.frames is a collection, not an array. We have to
  // perform this trick to make it work
  
  hahFrameArray = new Array();
  for (i = 0; i < win.frames.length; i++)
    hahFrameArray.push(win.frames[i]);

  while (win) {
    // move the area  a little bit to up-left,  so that sensible links
    // are shown
    area[0] = win.pageXOffset - 5;
    area[1] = win.pageYOffset - 5;
    area[2] = area[0] + win.innerWidth;
    area[3] = area[1] + win.innerHeight;
    counter += generateHints(win.document, counter, area);
//    hahLogMessage(win.document.location);
    win = getNextWin();
    if (win)
      for (i = 0; i < win.frames.length; i++)
	hahFrameArray.push(win.frames[i]);

    //    hahLogMessage(" " + hahDocCount);
  }

  hahStatusIcon("enabled", true);
}

function hahRemoveHints()
{
  for (var i = 0; i < hintElemArray.length; i++) {
    var hint = hintElemArray[i];
    hint.ownerDocument.body.removeChild(hint);
    //hintElemArray[i].parentNode.removeChild(hint);
  }
}

/**
 * Function generateHints
 *
 * Generates  hints for one  document, and  saves generated  hints and
 * corresponding  links  to  linkElemArray and  hintElemArray.   These
 * arrays  have to  be  initialized first,  in the  generateAllHints()
 * function.
 *
 * @author Pekka Sillanpaa
 * @param doc hints are generated for this document
 * @param firstId the first free id number for the linkElemArray and
 * hintElemArray
 * @return int number of generated elements
 */

function generateHints(doc, firstId, area)
{
  var linkCount = 0;
  var p = new point(0,0);

  var hintableElems = null;
  var elem = null, cs = null, hintElem = null;

  //  areCoordsValid = false;

  // go through the element list
  var tags = hahPrefs.get(HAH_PREFS.HINT_TAGS).split(';');
  var tag;

  if (tags.length == 0) {
    return linkCount;
  }

  for (tag = 0; tag < tags.length; tag++) {

    hintableElems = doc.getElementsByTagName(tags[tag]);

    for (i = 0; i < hintableElems.length; i++) {
      elem = hintableElems[i];
      
      if (elem.offsetWidth == 0 && elem.offsetHeight == 0)
	continue;
      
      cs = document.defaultView.getComputedStyle(elem, '');

      if (cs.visibility == "hidden")
	continue;

      // this has to be rethinked
      if (elem.tagName == 'A') {
	fc = elem.firstChild;
	if (fc != null) {
	  if (fc.tagName == 'IMG')
	    elem = fc;
	}
      }
       
      getElemCoords(elem, p);

      //if (p.y > area[3]) break; // all elems might not be shown, but this optimizes the speed in long pages
      //      if (p.y < area[1] || p.x > area[2] || p.x < area[0])
       if (p.y > area[3] || p.y < area[1] || p.x > area[2] || p.x < area[0])
      	continue;

      
      hintElem = makeHint(doc, formatHint(firstId + linkCount), firstId + linkCount, p.x, p.y);
      
      //elem.appendChild(hintElem);

      hintElemArray[firstId + linkCount] = hintElem; // save the hintelem so that it can be referenced
      linkElemArray[firstId + linkCount] = hintableElems[i];
      linkCount++;
    }
  }

  areCoordsValid = true;
  return linkCount;
}

function formatHint(hintNum) {
  var numbers = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  var str = hintNum.toString(numbers.length); // turn hintNum into a base(length) number
  
  if (hahPrefs.get(HAH_PREFS.HINT_FORCE_NUMERIC)) {
    return str;
  }
  else {
    // map the number onto the chars in the numbers string
    var result = '';
    for (var i=0; i<str.length; i++)
      result += (''+numbers.charAt(parseInt(str[i], numbers.length))).toUpperCase();;
    return result;
  }
}

function makeHint(doc, text, id, x, y)
{
  var styleString = hahPrefs.get(HAH_PREFS.HINT_STYLE);
  var hintElem = doc.createElement('SPAN');
  hintElem.appendChild(doc.createTextNode(text));
  hintElem.setAttribute('id', HINT_PREFIX + id);
  hintElem.setAttribute('style', styleString);
  hintElem.style.left = x + 'px';
  hintElem.style.top = y + 'px';
  hintElem.style.position = 'absolute';
  doc.body.appendChild(hintElem);
  return hintElem;
}

function getElemCoords(elem, p)
{
  var x = 0;
  var y = 0;
  var elemList = new Array();
  var elemTemp = elem;
  var parent = null;
  var j;

  // find the absolute coordinates
  while (parent != 'BODY') {


     // Position cacheing code ----
     // we need to ignore this if the page has been resized,
     // to force positions to be recalculated
    if (areCoordsValid)
    {
      if (elemTemp.absLeft != null) {
        x += elemTemp.absLeft;
        y += elemTemp.absTop;
	for (j = 0; j < elemList.length; j++) {
	  elemList[j].absLeft += elemTemp.absLeft;
          elemList[j].absTop += elemTemp.absTop;
        }
        break;
      }
    }
    elemTemp.absLeft = 0;
    elemTemp.absTop = 0;

    elemList[elemList.length] = elemTemp;
    for (j = 0; j < elemList.length; j++) {
      elemList[j].absLeft += elemTemp.offsetLeft;
      elemList[j].absTop += elemTemp.offsetTop;
    }

    x += elemTemp.offsetLeft;
    y += elemTemp.offsetTop;
    if (elemTemp.offsetParent != null)
      parent = elemTemp.offsetParent.tagName;
    else {
      hahLogMessage("ERROR: failed to define coordinates for link: " + elem);
      break;
    }
    elemTemp = elemTemp.offsetParent;
  }

  // find imagemap's coordinates
  if (elem.tagName == 'AREA') {
    var coords = elem.getAttribute("coords").split(",");
    x += Number(coords[0]);
    y += Number(coords[1]);
  }    

  p.x = x;
  p.y = y;
}

function point(x,y)
{
  this.x = x;
  this.y = y;
}

function getLinkElements()
{
  return linkElemArray;
}

function hahChangeHintFocus(linkNumString, oldLinkNumString)
{
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  var hintId = parseInt(linkNumString, nums.length);
  var oldHintId = parseInt(oldLinkNumString, nums.length);

  var styleString = hahPrefs.get(HAH_PREFS.HINT_STYLE);
  var styleStringFocus = hahPrefs.get(HAH_PREFS.HINT_STYLE_FOCUS);


  var xTemp, yTemp;

  if (oldLinkNumString.length > 0) {
    hintElem = hintElemArray[oldHintId];
    xTemp = hintElem.style.left;
    yTemp = hintElem.style.top;
    hintElem.setAttribute('style', styleString);
    hintElem.style.left = xTemp;
    hintElem.style.top = yTemp;
    hintElem.style.position = 'absolute';
  }

  hintElem = hintElemArray[hintId]; // TODO: has no properties??
  xTemp = hintElem.style.left;
  yTemp = hintElem.style.top;
  /*
    var cs = document.defaultView.getComputedStyle(hintElem, '');
    widthTemp = cs.width;
    heightTemp = cs.height;
  */
  hintElem.setAttribute('style', styleStringFocus);
  /*
    cs = document.defaultView.getComputedStyle(hintElem, '');
    widthTemp2 = cs.width;
    heightTemp2 = cs.height;
//    hahLogMessage((parseInt(xTemp) - (parseInt(widthTemp2) - parseInt(widthTemp)) / 2));
//    hahLogMessage(cs.height + " " + (parseInt(yTemp) - (parseInt(heightTemp2) - parseInt(heightTemp)) / 2));
  */
  hintElem.style.left = xTemp; //(parseInt(xTemp) - (parseInt(widthTemp2) - parseInt(widthTemp)) / 2) + "px";
  hintElem.style.top = yTemp; //(parseInt(yTemp) - (parseInt(heightTemp2) - parseInt(heightTemp)) / 2) + "px";
  hintElem.style.position = 'absolute';
}

function hahRegenHints()
{
  areCoordsValid = false;
}


////////////////////////////////////////////////////////////////////////////////
// basic functionality
////////////////////////////////////////////////////////////////////////////////

/**
 * Enables the HaH-mode by showing the hints and prepare for input the
 * hint numbers
 *
 * @author Pekka Sillanpaa
 * @param event that caused the mode to change
 * @return -1 if already enabled
 */

function hahEnableHahMode(event)
{
  if (isHahModeEnabled) 
    return -1;

  hahStatusIcon("wait", true); // why doesn't work? xul don't refresh?
  linkNumString = '';
  isHahModeEnabled = true;
  hahShowHints();

  hahFAYTState = gUseTypeAheadFind; // disable FAYT
  if (gUseTypeAheadFind) gUseTypeAheadFind = false;

  hahStatusIcon("active", true);
  return 0;
}

/**
 * Disables the HaH-mode hiding the hints and disabling the input mode
 *
 * @author Pekka Sillanpaa
 * @param event that caused the mode to change
 * @param action = true if something is to be clicked
 *                 false if cancel
 * @return -1 if already disabled
 */
function hahDisableHahMode(event, action)
{
  if (!isHahModeEnabled) 
    return -1;

  hahRemoveHints();
  hahStatusIcon("enabled", true);
  hahStatusDisplay("");
  isHahModeEnabled = false;
  
  var linksArray = getLinkElements();
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  var doc = window.document;
  
  var hintId = parseInt(linkNumString, nums.length);
  linkNumString = '';

  gUseTypeAheadFind = hahFAYTState; // re-enable FAYT if it was enabled

  if (hintId >= 0 && hintId < linksArray.length && action) {
    var evt = doc.createEvent('MouseEvents');
    var x = 0;
    var y = 0;
    // for imagemap
    if (linksArray[hintId].tagName == 'AREA') {
      var coords = linksArray[hintId].getAttribute("coords").split(",");
      x = Number(coords[0]);
      y = Number(coords[1]);
    }
    evt.initMouseEvent('click', true, true, doc.defaultView, 1, x, y, 0, 0, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, 0, null);
    linksArray[hintId].dispatchEvent(evt);

    // we don't want to remain the focus for radiobuttons and checkboxes
    if (linksArray[hintId].tagName == 'INPUT') {
      var type = linksArray[hintId].getAttribute("type");
      if (type == "radio" || type == "checkbox")
	hahBlurFocused();
    }
  } 
  return 0;
}

/**
 * Sets mouse over a link to enable the onmouseover event for that 
 * element
 *
 * @author Pekka Sillanpaa
 */
function hahSetMouseOverLink()
{
  var linksArray = getLinkElements();
  var doc = window.document;
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  
  var hintId = parseInt(linkNumString, nums.length);

  elem = linksArray[hintId];

  elem.focus();

  if (hintId >= 0 && hintId < linksArray.length) {
    var evt = doc.createEvent('MouseEvents');
    var x = 0;
    var y = 0;
    // for imagemap
    if (linksArray[hintId].tagName == 'AREA') {
      var coords = linksArray[hintId].getAttribute("coords").split(",");
      x = Number(coords[0]);
      y = Number(coords[1]);
    }
    evt.initMouseEvent('mouseover', true, true, doc.defaultView, 1, x, y, 0, 0, 0, 0, 0, 0, 0, null);
    linksArray[hintId].dispatchEvent(evt);
  }
}

// not necessarily needed anymore
function generateKeypress(keyCode, charCode, shiftKey)
{
  var doc = getFirstDoc();
  var evt = doc.createEvent('KeyEvents');
  evt.initKeyEvent('keypress', true, true, window, false, false, shiftKey, false, keyCode, charCode, doc.defaultView);
  var elt = document.commandDispatcher.focusedElement;
  if (elt != null) elt.dispatchEvent(evt);
  else doc.body.dispatchEvent(evt);
}

function charcodeToNumber(charCode)
{
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS).toLowerCase();
  var chr = String.fromCharCode(charCode).toLowerCase();
  var i = nums.indexOf(chr);
  if (i == -1) return null;
  return i.toString(nums.length);
}

function isFormElemFocused()
{
  if (hahPrefs.get(HAH_PREFS.KEYINPUT_IN_FORMS)) return false;

  var elt = document.commandDispatcher.focusedElement;
  if (elt == null) return false;

  var tagName = elt.localName.toUpperCase();

  //hahLogMessage(tagName);

  if (tagName == "INPUT" ||
      tagName == "TEXTAREA" ||
      tagName == "SELECT" ||
      tagName == "BUTTON" ||
      tagName == "ISINDEX")
    return true;
  
  return false;
}

function hahUpdateChanges(linkNumString, oldLinkNumString)
{
  //  hahStatusDisplay(linkNumString);
  hahChangeHintFocus(linkNumString, oldLinkNumString);
  hahSetMouseOverLink();
}

function hahTypeNumber(number)
{
  if (!isHahModeEnabled) return;

  var linksArray = getLinkElements();
  var oldLinkNumString = null; 
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  
  if (number != null) {
    oldLinkNumString = linkNumString;
    linkNumString += '' + number;
    var hintId = parseInt(linkNumString, nums.length);
    if (!(hintId >= 0 && hintId < linksArray.length))
      linkNumString = oldLinkNumString;
  }
 
  hahUpdateChanges(linkNumString, oldLinkNumString);
}

function hahTypeBackspace() 
{
  var oldLinkNumString = null;
  if (linkNumString.length > 0) {
    oldLinkNumString = linkNumString;
    linkNumString = linkNumString.substr(0, linkNumString.length - 1);
  }
  hahUpdateChanges(linkNumString, oldLinkNumString);
}

////////////////////////////////////////////////////////////////////////////////
// event handlers
////////////////////////////////////////////////////////////////////////////////

function hahOnKeydown(event)
{

  if (!document.getElementById("FindToolbar").hidden) return; // disable HaH if find toolbar is open

  if (event.keyCode == 27 && !isHahModeEnabled) {
    hahBlurFocused();
    return;
  }

  if (isFormElemFocused() && !event.altKey) return;
    
  if (hahIsKey(event, hahPrefs.get(HAH_PREFS.START_KEY))) {
    event.preventDefault();
    event.stopPropagation();

    if (!isHahModeEnabled) {
      hahEnableHahMode(event);
    }
    else {
      hahDisableHahMode(event, false);
    }
  }    
  else if (hahIsKey(event, hahPrefs.get(HAH_PREFS.MAGIC_KEY))) {
    hahEnableHahMode(event);
    isMagicKeyPressed = true;
  }
}

function hahOnKeyup(event)
{
  if (!isHahModeEnabled && isFormElemFocused()) return;

  var generate = false;
  if (hahIsKey(event, hahPrefs.get(HAH_PREFS.MAGIC_KEY))) {
    if (linkNumString.length == 0) generate = true;

    hahDisableHahMode(event, true);
    event.preventDefault();
    event.stopPropagation();
    // if nothing happened to the number string, most propably user wanted press the MAGIC_KEY in the focused element
    if (hahPrefs.get(HAH_PREFS.GENERATE_MAGIC) && generate) generateKeypress(0, hahPrefs.get(HAH_PREFS.MAGIC_KEY) >> 4, event.shiftKey);
    isMagicKeyPressed = false;
  }
}

function hahOnKeypress(event)
{
  if ((isFormElemFocused() && linkNumString == 0 && !event.altKey && event.keyCode != BACKSPACE_KEY) || !isHahModeEnabled) return;

  var num = charcodeToNumber(event.charCode);

  if (num != null) {
    event.preventDefault();
    event.stopPropagation(); 
    hahTypeNumber(num);
  }
  else if (event.keyCode == BACKSPACE_KEY && linkNumString.length > 0) {
    event.preventDefault();
    event.stopPropagation(); 
    hahTypeBackspace();    
    if (linkNumString == 0)
      hahDisableHahMode(event, false);
  } 
  else if (event.keyCode == STOP_KEY) {
    event.preventDefault();
    event.stopPropagation(); 
    hahDisableHahMode(event, false);
  }
  else if (isMagicKeyPressed) {
    event.preventDefault();
    event.stopPropagation(); 
  }
}

function hahIsKey(event, combinedKeyCode)
{
  var keyModifiers = combinedKeyCode & 0xF;
  var eventModifiers = event.altKey * Event.ALT_MASK |
                       event.ctrlKey * Event.CONTROL_MASK |
                       event.shiftKey * Event.SHIFT_MASK |
                       event.metaKey * Event.META_MASK;
  // we don't care about modifiers if it's a keyup event
  if (event.type.toLowerCase() == 'keyup') keyModifiers = eventModifiers;
  return ((event.keyCode == combinedKeyCode >> 4) &&
          (keyModifiers == eventModifiers));
  
}

////////////////////////////////////////////////////////////////////////////////
// statusbar manipulating functions
////////////////////////////////////////////////////////////////////////////////

function hahStatusDisplay(string)
{
  var hahStatus = document.getElementById('hah-status');
  var status;
  if (string.length > 0) {
    hahStatus.setAttribute("status", "active");
    hahStatus.hidden = false;
  } else {
    hahStatus.setAttribute("status", "inactive");
    hahStatus.hidden = true;
  }
  hahStatus.label = string;
}

function hahStatusIcon(status, enabled)
{
  var hahStatus = document.getElementById('hah-status-icon');
  if (!enabled || !hahPrefs.get(HAH_PREFS.STATUSICON_ENABLED)) {
    hahStatus.hidden = true;
  } else {
    hahStatus.hidden = false;
    hahStatus.setAttribute("status", status);
  }
}

function windowStatusDisplay(string)
{
  window.status = string;
}

window.document.addEventListener("load", hahLoad, true);
window.document.addEventListener("resize", hahResize, true);


//is there any method to do this?
//would be nice with js popup menus etc.
//window.document.addEventListener("mutation", hahMutate, true);
// temporal solution
window.document.addEventListener("click", hahResize, true);


hahLogMessage("Initialized.");

